from prettytable import PrettyTable
from colored import fg, bg, attr
from .core import main

import argparse

default_nodes = [
    "https://hived.emre.sh",
    "https://api.hive.blog",
]

ok_color = bg('green') + fg('white')
nok_color = bg('red') + fg('white')
reset = attr('reset')


def runner():
    parser = argparse.ArgumentParser()
    parser.add_argument('--nodes', nargs='+', )
    arguments = parser.parse_args()
    resp = main(nodes=arguments.nodes or default_nodes)

    for api_type, call in resp.items():
        table = PrettyTable()
        table.field_names = ["#", "Node", "Call", "Status", "Time [ms]",
                             "Error"]
        table.align = "l"
        for sub_call, scan_statuses in call.items():
            for scan_status, scan_results in scan_statuses.items():
                i = 1
                for scan_result in scan_results:
                    table.add_row(
                        [f"[{i}]",
                         scan_result["node"],
                         sub_call,
                         ok_color + "OK" + reset if scan_status else nok_color + "NOK" + reset,
                         int(scan_result["time_spent"] * 1e4),
                         scan_result.get("error", "-")]
                    )
                    i += 1
        print(table.get_string(title=f"{api_type}"))



from . import validators
from collections import namedtuple

fields = ['call', 'params', 'validator', 'validator_params']

check_definition = namedtuple(
    'CheckDefinition',
    fields,
)
check_definition.__new__.__defaults__ = (None,) * len(check_definition._fields)

CHECK_DEFINITIONS = [
    # transaction_status_api
    check_definition(
        "transaction_status_api.find_transaction",
        {"transaction_id": "b8660cc5464d37aa566e12d70349a30706a243063"},
    ),
    # account_by_key_api
    check_definition(
        "account_by_key_api.get_key_references",
        {},
    ),
    # condenser_api
    check_definition(
        "condenser_api.find_proposals",
        [[0]],  # proposal id: 0, return proposal.
        validators.should_return_at_least_one_element,
    ),
    check_definition(
        "condenser_api.get_account_count",
        [],
    ),
    check_definition(
        "condenser_api.get_account_history",
        ["hiveio", 100, 100],
    ),
    check_definition(
        "condenser_api.get_account_reputations",
        ["emre", 10],
        validators.should_return_at_least_one_element,
    ),
    check_definition(
        "condenser_api.get_accounts",
        [["emrebeyler"]],
        validators.should_return_at_least_one_element,
    ),
    check_definition(
        "condenser_api.get_active_witnesses",
        [],
        lambda resp: len(resp["result"]) == 21
    ),
    check_definition(
        "condenser_api.get_block",
        [8675309],
        lambda resp: resp["result"]["witness"] == "pfunk"
    ),
    check_definition(
        "condenser_api.get_block_header",
        [8675309],
        lambda resp: len(resp["result"].keys()) == 5,
    ),
    check_definition(
        "condenser_api.get_blog",
        ["hiveio", 0, 1],
        lambda resp: len(resp["result"]) == 1
    ),
    check_definition(
        "condenser_api.get_blog_entries",
        ["hiveio", 0, 3],
    ),
    check_definition(
        "condenser_api.get_chain_properties",
        [],
        lambda resp: isinstance(resp.get("result"), dict)
    ),
    check_definition(
        "condenser_api.get_comment_discussions_by_payout",
        [{"tag": "hive", "limit": 1}],
        validators.should_return_at_least_one_element,
    ),
    check_definition(
        "condenser_api.get_config",
        [],
        validators.should_return_a_dictionary,
    ),
    check_definition(
        "condenser_api.get_content",
        ["emrebeyler", "httpx"],
        validators.should_return_a_dictionary,
    ),
    check_definition(
        "tags_api.get_content_replies",
        {"author": "emrebeyler", "permlink": "httpx"},
        validators.should_return_at_least_one_element,
    ),
    check_definition(
        "block_api.get_block",
        {"block_num": 8675309},
        lambda resp: resp.get("result", {}).get("block", {}).get(
            "witness") == 'pfunk'
    ),
    check_definition(
        "block_api.get_block_header",
        {"block_num": 8675309},
        lambda resp: resp.get("result", {}).get("header", {}).get(
            "witness") == 'pfunk'
    ),
    check_definition(
        "database_api.get_current_price_feed",
        {"block_num": 8675309},
        validators.should_return_a_dictionary,
    ),
    check_definition(
        "database_api.get_dynamic_global_properties",
        {},
        validators.should_return_a_dictionary,
    ),

]



import httpx
import asyncio
import uuid
import time
import traceback

from .checkers import CHECK_DEFINITIONS


def build_request_body(method, params):
    """Build a JSON-RPC request body based on the parameters given."""
    data = {
        "jsonrpc": "2.0",
        "method": method,
        "params": params,
        "id": str(uuid.uuid4())
    }
    return data


async def rpc_request(client, node, call, params, validator):
    response = await client.post(
        node,
        json=build_request_body(call, params)
    )
    response = response.json()

    status = True
    if 'error' in response:
        status = False
    else:
        if validator:
            status = validator(response)
            if not status:
                # log the resp here
                # since the validation is failed
                print(response)

    return status, response


async def runner(nodes):
    results = {}
    total_requests = 0
    async with httpx.AsyncClient() as client:
        for definition in CHECK_DEFINITIONS:
            api_type = definition.call.split(".")[0]
            for node in nodes:
                result = {"node": node}
                start_time = time.time()
                status = None
                try:
                    status, response = await rpc_request(
                        client, node, definition.call,
                        definition.params, definition.validator)
                    if not status:
                        # in case we have a RPC error
                        # update the dict with the response dict
                        # so that we can investigate the error details
                        rpc_error = response.get(
                            "error", {}).get("message")
                        if rpc_error:
                            result["error"] = rpc_error
                except Exception as error:
                    traceback.print_exc()
                    result.update({"error": error.args[0]})
                end_time = time.time()
                result.update({"time_spent": round(end_time - start_time, 4)})
                total_requests += 1

                # boiler plate checks to set the default dict
                if api_type not in results:
                    results[api_type] = {}

                if definition.call not in results[api_type]:
                    results[api_type][definition.call] = {}

                if status not in results[api_type][definition.call]:
                    results[api_type][definition.call][status] = []

                results[api_type][definition.call][status].append(result)
            results[api_type][definition.call][status].sort(
                key=lambda x: x["time_spent"])

    return results, total_requests


def main(nodes):
    start_time = time.time()
    loop = asyncio.get_event_loop()
    results, total_requests  =loop.run_until_complete(runner(nodes))
    end_time = time.time()

    print(f" > {total_requests} "
          f"requests sent in {round(end_time - start_time, 2)} seconds.")
    return results



MIN_BLOCKCHAIN_VERSION = 1244  # 1.24.4
CHAIN_ID = 'beeab0de00000000000000000000000000000000000000000000000000000000'


def should_return_at_least_one_element(response, response_key="result"):
    results = response.get(response_key)
    if not isinstance(results, list):
        return False

    if len(results) == 0:
        print(response)
        return False

    return True


def should_return_a_dictionary(response, response_key="result"):
    results = response.get(response_key)
    if not isinstance(results, dict):
        return False

    if not len(results.keys()):
        return False

    return True


def validate_get_version(response):
    version = response.get("result", {}).get("blockchain_version")
    if not version:
        return False
    if int(version.replace(".", "")) < MIN_BLOCKCHAIN_VERSION:
        return False

    if response.get("result", {}).get("chain_id") < CHAIN_ID:
        return False

    return True



from setuptools import setup

setup(
    name='hived_rpc_scanner',
    version='0.0.6',
    packages=["hived_rpc_scanner",],
    url='https://github.com/emre/hived-rpc-scanner',
    license='MIT',
    author='emre yilmaz',
    author_email='mail@emreyilmaz.me',
    description='A tool to check status of certain endpoints on Hived RPC',
    entry_points={
        'console_scripts': [
            'hived_rpc_scanner = hived_rpc_scanner.runner:runner',
        ],
    },
    install_requires=["prettytable==2.0.0", "httpx", "colored"]
)